from attrs import define, field
from abc import ABC, abstractmethod
from typing import Optional
from galaxybrain.prompts.prompt import Prompt
from galaxybrain.completions.completion_result import CompletionResult
from galaxybrain.memory import Memory


@define
class Completion(ABC):
    memory: Optional[Memory] = field(default=None, kw_only=True)

    def __call__(self, prompt: any, memory: Optional[Memory] = None) -> CompletionResult:
        if isinstance(prompt, Prompt):
            return self.complete(prompt, memory)
        elif isinstance(prompt, str):
            return self.complete(Prompt(prompt, memory))
        else:
            raise Exception("Unknown prompt type")

    @abstractmethod
    def complete(self, prompt: Prompt, memory: Optional[Memory] = None) -> CompletionResult:
        pass

    


from attrs import define
from typing import Callable
from galaxybrain.completions.completion_result import CompletionResult


@define
class Rule():
    value: str
    validator: Callable[[CompletionResult], bool] = lambda v: True


from galaxybrain.rules.rule import Rule
from galaxybrain.completions.completion_result import CompletionResult
import json


def return_valid_json() -> bool:
    return Rule(
        "only output valid JSON",
        validator=lambda result : __is_valid_json(result.value)
    )

def return_array() -> bool:
    return Rule(
        "only output a valid JSON array",
        validator=lambda result : __is_json_array(result.value)
    )

def return_object() -> bool:
    return Rule(
        "only output a valid JSON object",
        validator=lambda result : __is_json_object(result.value)
    )

def put_answer_in_field(value: str) -> bool:
    return Rule(
        f"only output a valid JSON object with your answer in '{value}'",
        validator=lambda result : __has_key_in_json_object(result.value, value)
    )

def __is_valid_json(json_string: str) -> bool:
    try:
        json.loads(json_string)

        return True
    except json.decoder.JSONDecodeError:
        return False

def __is_json_array(string):
    try:
        json_data = json.loads(string)

        if isinstance(json_data, list):
            return True
        else:
            return False
    except:
        return False

def __is_json_object(string):
    try:
        json_data = json.loads(string)

        if isinstance(json_data, dict):
            return True
        else:
            return False
    except:
        return False

def __has_key_in_json_object(json_string: str, key: str) -> bool:
    try:
        json_data = json.loads(json_string)

        if key in json_data:
            return True
        return False
    except json.decoder.JSONDecodeError:
        return False

def __has_keys_in_json_array(json_string: str, key: str) -> bool:
    try:
        json_data = json.loads(json_string)

        return all([__has_key_in_json_object(json.dumps(json_obj), key) for json_obj in json_data])
    except json.decoder.JSONDecodeError:
        return False


from attrs import define, field
from typing import Optional
from galaxybrain.rules.rule import Rule
from galaxybrain.memory import Memory
from jinja2 import Environment, FileSystemLoader
import galaxybrain
import os


@define
class Prompt():
    value: str
    rules: list[Rule] = field(default=[], kw_only=True)
    memory: Optional[Memory] = field(default=None, kw_only=True)
    j2_env: any = field(init=False)

    def __attrs_post_init__(self):
        templates_path = os.path.join(galaxybrain.PACKAGE_ABS_PATH, "prompts/templates")

        self.j2_env = Environment(
            loader=FileSystemLoader(templates_path),
            trim_blocks=True,
            lstrip_blocks=True
        )

    def build(self) -> str:
        if self.memory:
            if self.memory.is_empty():
                prompt = self.intro_prompt()

                self.memory.add_memory(prompt)
            else:
                prompt = self.memory_prompt()

                self.memory.add_memory(self.format_answer(self.value))
        else:
            prompt = self.intro_prompt()

        return prompt
    
    def validate(self, value: str) -> bool:
        return all(rule.validator(value) for rule in self.rules)

    def intro_prompt(self) -> str:
        rules_text = self.j2_env.get_template("rules.j2").render(
            {
                "rules": self.rules
            }
        )

        intro_text = self.j2_env.get_template("intro.j2").render(
            {
                "question": self.format_question(self.value),
                "answer": self.format_answer("")
            },
        )

        return f"{rules_text}\n{intro_text}"
    
    def memory_prompt(self) -> str:
        return self.j2_env.get_template("memory.j2").render(
            {
                "memory": self.memory.to_string(),
                "question": self.format_question(self.value),
                "answer": self.format_answer("")
            }
        )

    def format_question(self, question: str) -> str:
        return f"Q: {question}"
    
    def format_answer(self, answer: str) -> str:
        return f"A: {answer}"


from attrs import define

from galaxybrain.completions.completion import Completion
from galaxybrain.prompts.prompt import Prompt
from galaxybrain.completions.completion_result import CompletionResult

import openai
import json


@define()
class OpenAiCompletion(Completion):
    api_key: str = None
    model: str = "text-davinci-003"
    suffix: str = None
    max_tokens: int = 2000
    temperature: float = 0.5
    top_p: float = 1
    n: int = 1
    stream: bool = False
    logprobs: int = None
    echo: bool = False
    stop = None
    presence_penalty: int = 0
    frequency_penalty: int = 0
    best_of: int = 1
    logit_bias: map = {}
    user: str = ""

    def complete(self, prompt: Prompt) -> CompletionResult:
        if self.api_key:
            openai.api_key = self.api_key

        result = openai.Completion.create(
            model=self.model,
            prompt=prompt.build(),
            suffix=self.suffix,
            max_tokens=self.max_tokens,
            temperature=self.temperature,
            top_p=self.top_p,
            n=self.n,
            stream=self.stream,
            logprobs=self.logprobs,
            echo=self.echo,
            stop=self.stop,
            presence_penalty=self.presence_penalty,
            frequency_penalty=self.frequency_penalty,
            best_of=self.best_of,
            logit_bias=self.logit_bias,
            user=self.user
        )

        if len(result.choices) == 1:
            result_choice = result.choices[0].text.strip()

            if prompt.memory:
                prompt.memory.add_memory(result_choice)

            return CompletionResult(
                value=result_choice,
                meta={
                    "id": result["id"],
                    "created": result["created"],
                    "usage": json.dumps(result["usage"])
                }
            )
        else:
            raise Exception("Completion with more than one choice is not supported yet.")


from attrs import define
from typing import Optional
from galaxybrain.completions.completion_result import CompletionResult
from galaxybrain.rules.rule import Rule


@define()
class Validator():
    result: CompletionResult
    rules: list[Rule]
    rule_validations: dict[Rule, bool] = []

    def is_valid(self) -> Optional[bool]:
        if len(self.rules) == len(self.rule_validations):
            return all([validation[1] for validation in self.rule_validations])
        else:
            None

    def validate(self) -> bool:
        self.rule_validations = [(rule, rule.validator(self.result)) for rule in self.rules]

        return self.is_valid()

    def failed_rules(self) -> list[Rule]:
        return [validation[0].value for validation in self.rule_validations if not validation[1]]


from attrs import define

@define
class Memory:
    memories: list[str] = []

    def is_empty(self) -> bool:
        return len(self.memories) == 0
    
    def add_memory(self, memory: str) -> None:
        return self.memories.append(memory)
    
    def to_string(self):
        return "\n".join(self.memories)


from galaxybrain.completions.completion import Completion
from galaxybrain.completions.completion_result import CompletionResult
from galaxybrain.completions.openai_completion import OpenAiCompletion
from galaxybrain.completions.validator import Validator


__all__ = [
    "Completion",
    "OpenAiCompletion",
    "CompletionResult",
    "Validator"
]


import pathlib
import galaxybrain
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="galaxybrain",
    version=galaxybrain.VERSION,
    description="LLM extensions framework",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/vasinov/galaxybrain",
    author="Vasily Vasinov",
    author_email="vasinov@me.com",
    license="Apache 2.0",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "python-dotenv>=0.21",
        "openai>=0.26",
        "attrs>=22",
        "jinja2>=3.1"
    ]
)